#ifndef NPSI_MINUITDENSITYFITFCN1D_HH_
#define NPSI_MINUITDENSITYFITFCN1D_HH_

/*!
// \file MinuitDensityFitFcn1D.hh
//
// \brief Binned fit of a parametric density by maximum likelihood
//
// Author: I. Volobouev
//
// September 2010
*/

#include <cfloat>

#include "npstat/stat/HistoND.hh"
#include "npstat/stat/FitUtils.hh"

#include "Minuit2/FCNBase.h"

namespace npsi {
    /**
    // Target minimization function adapter class for running maximum
    // likelihood density fits to histogrammed data by Minuit2.
    //
    // DensityConstructor is a functor which creates the necessary
    // density function on the stack out of vector of parameters.
    // Must have "operator()(const std::vector<double>&) const"
    // which returns an object (or a reference) of some class which
    // was derived from AbsDistribution1D.
    */
    template <class Numeric, class DensityConstructor>
    class MinuitDensityFitFcn1D : public ROOT::Minuit2::FCNBase
    {
    public:
        /** This class will not assume ownership of any pointers or references */
        MinuitDensityFitFcn1D(const npstat::HistoND<Numeric>& histo,
                              const unsigned char* binMask,
                              const unsigned maskLength,
                              const DensityConstructor& densityMaker,
                              const double minlog=log(DBL_MIN),
                              const double up=0.05)
            : histo_(histo),
              binMask_(binMask),
              densityMaker_(densityMaker),
              workspace_(maskLength),
              minlog_(minlog),
              up_(up)
        {
            assert(binMask);
            assert(maskLength);
            assert(histo_.dim() == 1U);
            assert(histo_.nBins() == maskLength);
        }

        virtual ~MinuitDensityFitFcn1D() {}

        /** This method returns negative log likelihood */
        inline double operator()(const std::vector<double>& x) const
        {
            return -npstat::densityFitLogLikelihood1D(
                histo_, binMask_, histo_.nBins(), densityMaker_(x),
                minlog_, &workspace_[0], workspace_.size());
        }

        inline double Up() const {return up_;}

    private:
        const npstat::HistoND<Numeric>& histo_;
        const unsigned char* binMask_;
        const DensityConstructor& densityMaker_;
        mutable std::vector<double> workspace_;
        double minlog_;
        double up_;
    };
}

#endif // NPSI_MINUITDENSITYFITFCN1D_HH_
