#ifndef NPSTAT_GAUSSLEGENDREQUADRATURE_HH_
#define NPSTAT_GAUSSLEGENDREQUADRATURE_HH_

/*!
// \file GaussLegendreQuadrature.hh
//
// \brief Gauss-Legendre quadratures in long double precision
//
// Author: I. Volobouev
//
// April 2010
*/

#include <vector>
#include <utility>

#include "npstat/nm/SimpleFunctors.hh"

namespace npstat {
    /** 
    // Gauss-Legendre quadrature. Internally, operations are performed
    // in long double precision.
    */
    class GaussLegendreQuadrature
    {
    public:
        /**
        // At the moment, the following numbers of points
        // are supported: 2, 4, 6, 8, 10, 12, 16, 32, 64, 100, 128, 256.
        //
        // If an unsupported number of points is given in the
        // constructor, std::invalid_argument exception will be thrown.
        */
        explicit GaussLegendreQuadrature(unsigned npoints);

        /** Return the number of quadrature points */
        inline unsigned npoints() const {return npoints_;}

        /**
        // The abscissae are returned for positive points only,
        // so the buffer length should be at least npoints/2.
        */
        void getAbscissae(long double* abscissae, unsigned len) const;

        /**
        // The weights are returned for positive points only,
        // so the buffer length should be at least npoints/2.
        */
        void getWeights(long double* weights, unsigned len) const;

        /** Perform the quadrature */
        template <typename FcnResult, typename FcnArg>
        long double integrate(const Functor1<FcnResult,FcnArg>& fcn,
                              long double a, long double b) const;

        /**
        // This method splits the interval [a, b] into "nsplit"
        // subintervals of equal length, applies Gauss-Legendre
        // quadrature to each subinterval, and sums the results.
        */
        template <typename FcnResult, typename FcnArg>
        long double integrate(const Functor1<FcnResult,FcnArg>& fcn,
                              long double a, long double b,
                              unsigned nsplit) const;
    private:
        GaussLegendreQuadrature();

        const long double* a_;
        const long double* w_;
        mutable std::vector<std::pair<long double, long double> > buf_;
        unsigned npoints_;

#ifdef SWIG
    public:
        inline std::vector<double> abscissae2() const
        {
            return std::vector<double>(a_, a_+npoints_/2U);
        }

        inline std::vector<double> weights2() const
        {
            return std::vector<double>(w_, w_+npoints_/2U);
        }
#endif
    };
}

#include "npstat/nm/GaussLegendreQuadrature.icc"

#endif // NPSTAT_GAUSSLEGENDREQUADRATURE_HH_
