#ifndef NPSTAT_MATRIX_HH_
#define NPSTAT_MATRIX_HH_

/*!
// \file Matrix.hh
//
// \brief Template matrix class
//
// Author: I. Volobouev
//
// November 2008
*/

#include "geners/ClassId.hh"
#include "npstat/nm/EigenMethod.hh"

namespace npstat {
    /**
    // A simple helper class for matrix manipulations. Depending on how much
    // space is provided with the "Len" parameter, the data will be placed
    // either on the stack or on the heap.
    //
    // C storage convention is used for internal data. In the element access
    // operations, array bounds are not checked.
    //
    // Note that this class is slower than matrix classes based on expression
    // templates (such as those in boost uBLAS or in Blitz++). If speed is
    // really important for your calculations, consider using a dedicated
    // matrix library.
    */
    template<typename Numeric, unsigned Len=16>
    class Matrix
    {
        template <typename Num2, unsigned Len2>
        friend class Matrix;

    public:
        typedef Numeric value_type;

        /**
        // Default constructor creates an unitialized matrix
        // which can be assigned from other matrices
        */
        Matrix();

        /**
        // This constructor creates an unitialized matrix
        // which can be assigned element-by-element or from
        // another matrix with the same dimensions
        */
        Matrix(unsigned nrows, unsigned ncols);

        /**
        // This constructor initializes the matrix as follows:
        //
        //   initCode = 0:  all elements are initialized to 0.
        //
        //   initCode = 1:  matrix must be square; diagonal elements are
        //                  initialized to 1, all other elements to 0.
        */
        Matrix(unsigned nrows, unsigned ncols, int initCode);

        /**
        // This constructor initializes the matrix from the
        // given 1-d array using C storage conventions
        */
        Matrix(unsigned nrows, unsigned ncols, const Numeric* data);

        /** Copy constructor */
        Matrix(const Matrix&);

        /** Converting copy constructor */
        template <typename Num2, unsigned Len2>
        Matrix(const Matrix<Num2, Len2>&);

        ~Matrix();

        /** 
        // Assignment operator. The matrix on the left must either be
        // in an uninitialized state or have compatible dimensions with
        // the matrix on the right.
        */
        Matrix& operator=(const Matrix&);

        /** Converting assignment operator */
        template <typename Num2, unsigned Len2>
        Matrix& operator=(const Matrix<Num2, Len2>&);

        //@{
        /** A simple inspection of matrix properties */
        inline unsigned nRows() const {return nrows_;}
        inline unsigned nColumns() const {return ncols_;}
        inline unsigned length() const {return nrows_*ncols_;}
        inline Numeric* data() const {return data_;}
        inline bool isSquare() const {return data_ && ncols_ == nrows_;}
        bool isSymmetric() const;
        bool isAntiSymmetric() const;
        //@}

        /** This method resets the object to an unintialized state */
        void reset();

        /**
        // This method changes the object dimensions.
        // All data is lost in the process.
        */
        void resize(unsigned nrows, unsigned ncols);

        /** This method sets all elements to 0 */
        void zeroOut();

        //@{
        /** Compare two matrices for equality */
        bool operator==(const Matrix&) const;
        bool operator!=(const Matrix&) const;
        //@}

        /**
        // Non-const access to the data (works like 2-d array in C).
        // No bounds checking.
        */
        Numeric* operator[](unsigned) const;

        /** Data modification method with bounds checking */
        void set(unsigned row, unsigned column, Numeric value);

        /** Access by value without bounds checking */
        Numeric operator()(unsigned row, unsigned column) const;

        /** Access by value with bounds checking */
        Numeric at(unsigned row, unsigned column) const;

        //@{
        /** Binary algebraic operations with matrices and vectors */
        Matrix operator*(const Matrix& r) const;
        Matrix operator*(Numeric r) const;
        Matrix operator/(Numeric r) const;
        Matrix operator+(const Matrix& r) const;
        Matrix operator-(const Matrix& r) const;
        Matrix operator+() const;
        Matrix operator-() const;
        //@}

        //@{
        /** In-place algebraic operations with matrices and vectors */
        Matrix& operator*=(Numeric r);
        Matrix& operator/=(Numeric r);
        Matrix& operator+=(const Matrix& r);
        Matrix& operator-=(const Matrix& r);
        //@}

        /** Multiplication by a vector represented by a pointer and array size */
        template <typename Num2>
        Matrix timesVector(const Num2* data, unsigned dataLen) const;

        /** Bilinear form with a vector (v^T M v) */
        template <typename Num2>
        Numeric bilinear(const Num2* data, unsigned dataLen) const;

        /**
        // Solution of a linear system of equations M*x = rhs.
        // The matrix must be square. "false" is returned in case
        // the matrix is degenerate.
        */
        bool solveLinearSystem(const Numeric* rhs, unsigned lenRhs,
                               Numeric* solution) const;

        /** Return transposed matrix */
        Matrix T() const;

        /** Construct a symmetric matrix from this one (symmetrize) */
        Matrix symmetrize() const;

        /** Construct an antisymmetric matrix from this one (antisymmetrize) */
        Matrix antiSymmetrize() const;

        /** Maximum absolute value for any row/column */
        Numeric maxAbsValue() const;

        //@{
        /** Calculate the trace (matrix must be square) */
        Numeric tr() const;
        inline Numeric sp() const {return tr();}
        //@}

        /** Calculate the determinant (matrix must be square) */
        Numeric det() const;

        /**
        // Inverse of a real symmetric positive definite matrix. Use this
        // function (it has better numerical precision than symInv()) if
        // you know that your matrix is symmetric and positive definite.
        // For example, a non-singular covariance matrix calculated for
        // some dataset has these properties.
        */
        Matrix symPDInv() const;

        /**
        // Inverse of a real symmetric matrix. If the matrix is not symmetric,
        // its symmetrized version will be used internally. Use this function
        // (it has better numerical precision than inv()) if you know that
        // your matrix is symmetric. 
        */
        Matrix symInv() const;

        /**
        // Inverse of a general matrix (which must be square and non-singular)
        */
        Matrix inv() const;

        /**
        // Eigenvalues and eigenvectors of a real symmetric matrix.
        // The "eigenvectors" pointer can be NULL if only eigenvalues
        // are needed. If it is not NULL, the _rows_ of that matrix
        // will be set to the computed eigenvectors.
        //
        // The "m" parameter specifies the LAPACK method to use for
        // calculating eigenvalues.
        */
        void symEigen(Numeric* eigenvalues, unsigned lenEigenvalues,
                      Matrix* eigenvectors, EigenMethod m=EIGEN_SIMPLE) const;

        /**
        // Calculate entropy-based effective rank. It is assumed that
        // the matrix is symmetric positive semidefinite.
        //
        // The "tol" parameter specifies the eigenvalue cutoff:
        // all eigenvalues equal to tol*maxEigenvalue or smaller
        // will be ignored. This parameter must be non-negative.
        //
        // The "m" parameter specifies the LAPACK method to use for
        // calculating eigenvalues.
        //
        // If "eigenSum" pointer is not NULL, *eigenSum will be filled
        // on exit with the sum of eigenvalues above the cutoff.
        */
        double symPSDefEffectiveRank(double tol = 0.0,
                                     EigenMethod m = EIGEN_D_AND_C,
                                     double* eigenSum = 0) const;

        /**
        // Calculate a function of this matrix. It is assumed that the
        // matrix is symmetric and real. The calculation is performed by
        // calculating the eignenbasis, evaluating the function of the
        // eigenvalues, and then transforming back to the original basis.
        */
        template <class Functor>
        Matrix symFcn(const Functor& fcn) const;

        /** Power of a matrix. Matrix must be square. */
        Matrix pow(unsigned degree) const;

        /**
        // The following function derives a correlation matrix
        // out of a covariance matrix. In the returned matrix
        // all diagonal elements will be set to 1.
        */
        Matrix covarToCorr() const;

        //@{
        /** Method related to "geners" I/O */
        inline gs::ClassId classId() const {return gs::ClassId(*this);}
        bool write(std::ostream& of) const;
        //@}

        static const char* classname();
        static inline unsigned version() {return 1;}
        static void restore(const gs::ClassId& id, std::istream& in, Matrix* m);

    private:
        Numeric local_[Len];
        Numeric* data_;
        unsigned nrows_;
        unsigned ncols_;
        unsigned len_;

#ifdef SWIG
    public:
        inline Matrix timesVector2(const Numeric* data, unsigned dataLen) const
            {return timesVector(data, dataLen);}

        inline Numeric bilinear2(const Numeric* data, unsigned dataLen) const
            {return bilinear(data, dataLen);}

        inline std::vector<Numeric> solveLinearSystem2(
            const Numeric* data, unsigned dataLen) const
        {
            std::vector<Numeric> result(dataLen);
            if (!solveLinearSystem(data, dataLen, dataLen ? &result[0] : 0))
                result.clear();
            return result;
        }

        inline std::vector<Numeric> symEigenValues() const
        {
            std::vector<Numeric> result(nrows_);
            symEigen(&result[0], nrows_, 0);
            return result;
        }
#endif
    };
}

#include <iostream>

template<typename N, unsigned Len>
std::ostream& operator<<(std::ostream& os, const npstat::Matrix<N, Len>& m);

#include "npstat/nm/Matrix.icc"

#endif // NPSTAT_MATRIX_HH_
