#include <cmath>

#include "npstat/nm/allocators.hh"
#include "npstat/nm/OrthoPoly1D.hh"

namespace npstat {
    OrthoPoly1D::OrthoPoly1D()
        : weight_(0), poly_(0), step_(0.0L), nw_(0), maxdeg_(0)
    {
    }

    OrthoPoly1D::OrthoPoly1D(const OrthoPoly1D& r)
        : weight_(0), poly_(0), step_(r.step_), nw_(r.nw_), maxdeg_(r.maxdeg_)
    {
        if (nw_)
        {
            weight_ = new long double[nw_];
            poly_ = new long double[nw_*(maxdeg_+1)];
            copyBuffer(weight_, r.weight_, nw_);
            copyBuffer(poly_, r.poly_, nw_*(maxdeg_+1));
        }
    }

    long double OrthoPoly1D::scalarProduct(
        const long double *x, const long double *y) const
    {
        long double sum = 0.0L;
        for (unsigned i=0; i<nw_; ++i)
            sum += x[i]*y[i]*weight_[i];
        return sum*step_;
    }

    long double OrthoPoly1D::scalarProduct(
        const double *x, const long double *y) const
    {
        long double sum = 0.0L;
        for (unsigned i=0; i<nw_; ++i)
            sum += x[i]*y[i]*weight_[i];
        return sum*step_;
    }

    void OrthoPoly1D::gramSchmidt(const unsigned startingDegree)
    {
        for (unsigned ideg=startingDegree; ideg<=maxdeg_; ++ideg)
        {
            long double *data_deg = poly_ + ideg*nw_;
            for (unsigned i=0; i<ideg; ++i)
            {
                const long double *data_i = poly_ + i*nw_;
                const long double norm = scalarProduct(data_deg, data_i);
                for (unsigned j=0; j<nw_; ++j)
                    data_deg[j] -= norm*data_i[j];
            }
            long double norm = scalarProduct(data_deg, data_deg);
            assert(norm > 0.0L);
            norm = sqrtl(norm);
            for (unsigned i=0; i<nw_; ++i)
                data_deg[i] /= norm;
        }
    }

    OrthoPoly1D::OrthoPoly1D(const unsigned maxDeg, const double* weight,
                             const unsigned weightLen, const double istep)
        : weight_(0), poly_(0), step_(istep), nw_(weightLen), maxdeg_(maxDeg)
    {
        if (!nw_ || nw_ <= maxdeg_) throw std::invalid_argument(
            "In npstat::OrthoPoly1D constructor: "
            "insufficient number of elements in the weight array");
        if (step_ <= 0.0) throw std::invalid_argument(
            "In npstat::OrthoPoly1D constructor: "
            "step size must be positive");
        assert(weight);
        bool havePositive = false;
        for (unsigned i=0; i<nw_; ++i)
        {
            if (weight[i] < 0.0)
                throw std::invalid_argument(
                    "In npstat::OrthoPoly1D constructor: "
                    "all weights must be non-negative");
            if (weight[i] > 0.0)
                havePositive = true;
        }
        if (!havePositive) throw std::invalid_argument(
            "In npstat::OrthoPoly1D constructor: no positive weights");

        weight_ = new long double[nw_];
        for (unsigned i=0; i<nw_; ++i)
            weight_[i] = weight[i];

        poly_ = new long double[nw_*(maxdeg_+1)];

        // Generate the 0th order polynomial
        for (unsigned i=0; i<nw_; ++i)
            poly_[i] = 1.0L;

        // Normalize the weight
        const long double norm = scalarProduct(poly_, poly_);
        for (unsigned i=0; i<nw_; ++i)
            weight_[i] /= norm;

        // Generate the monomials
        for (unsigned ideg=1; ideg<=maxDeg; ++ideg)
        {
            long double *data = poly_ + ideg*nw_;
            for (unsigned i=0; i<nw_; ++i)
                data[i] = powl(i*istep, ideg);
        }

        // Apply the Gram-Schmidt process. Repeat it to improve orthogonality.
        gramSchmidt(1);
        gramSchmidt(1);
    }

    OrthoPoly1D::~OrthoPoly1D()
    {
        delete [] poly_;
        delete [] weight_;
    }

    OrthoPoly1D& OrthoPoly1D::operator=(const OrthoPoly1D& r)
    {
        if (this == &r)
            return *this;
        delete [] poly_; poly_ = 0;
        delete [] weight_; weight_ = 0;
        step_ = r.step_;
        nw_ = r.nw_;
        maxdeg_ = r.maxdeg_;
        if (nw_)
        {
            weight_ = new long double[nw_];
            poly_ = new long double[nw_*(maxdeg_+1)];
            copyBuffer(weight_, r.weight_, nw_);
            copyBuffer(poly_, r.poly_, nw_*(maxdeg_+1));
        }
        return *this;
    }

    double OrthoPoly1D::weight(const unsigned ix) const
    {
        if (!weight_) throw std::invalid_argument(
            "In npstat::OrthoPoly1D::weight: uninitilalized polynomial set");
        if (ix >= nw_) throw std::out_of_range(
            "In npstat::OrthoPoly1D::weight: point index out of range");
        return weight_[ix];
    }

    double OrthoPoly1D::poly(const unsigned deg, const unsigned ix) const
    {
        if (!poly_) throw std::invalid_argument(
            "In npstat::OrthoPoly1D::poly: uninitilalized polynomial set");
        if (deg > maxdeg_) throw std::out_of_range(
            "In npstat::OrthoPoly1D::poly: poly degree out of range");
        if (ix >= nw_) throw std::out_of_range(
            "In npstat::OrthoPoly1D::poly: point index out of range");
        return poly_[deg*nw_ + ix];
    }

    double OrthoPoly1D::series(const double *coeffs,
                               const unsigned maxdeg,
                               const unsigned ux) const
    {
        if (!poly_) throw std::invalid_argument(
            "In npstat::OrthoPoly1D::series: uninitilalized polynomial set");
        if (maxdeg > maxdeg_) throw std::out_of_range(
            "In npstat::OrthoPoly1D::series: filter degree out of range");
        if (ux >= nw_) throw std::out_of_range(
            "In npstat::OrthoPoly1D::series: point index out of range");
        assert(coeffs);

        long double sum = 0.0L;
        const long double* localPoly = poly_ + ux;
        for (unsigned deg=0; deg <= maxdeg; ++deg)
            sum += coeffs[deg]*localPoly[deg*nw_];
        return sum;
    }

    void OrthoPoly1D::calculateCoeffs(const double *data,
                                      const unsigned dataLen,
                                      double *coeffs,
                                      const unsigned maxdeg) const
    {
        if (!weight_) throw std::invalid_argument(
            "In npstat::OrthoPoly1D::calculateCoeffs: "
            "uninitilalized polynomial set");
        if (dataLen != nw_) throw std::invalid_argument(
            "In npstat::OrthoPoly1D::calculateCoeffs: "
            "incompatible data length");
        if (maxdeg > maxdeg_) throw std::out_of_range(
            "In npstat::OrthoPoly1D::calculateCoeffs: "
            "filter degree out of range");
        assert(data);
        assert(coeffs);

        for (unsigned deg=0; deg<=maxdeg; ++deg)
            coeffs[deg] = scalarProduct(data, poly_ + deg*nw_);
    }

    bool OrthoPoly1D::operator==(const OrthoPoly1D& r) const
    {
        if (step_ != r.step_ || nw_ != r.nw_ || maxdeg_ != r.maxdeg_)
            return false;
        for (unsigned i=0; i<nw_; ++i)
            if (weight_[i] != r.weight_[i])
                return false;
        const unsigned len = (maxdeg_+1)*nw_;
        for (unsigned i=0; i<len; ++i)
            if (poly_[i] != r.poly_[i])
                return false;
        return true;
    }

    bool OrthoPoly1D::operator!=(const OrthoPoly1D& r) const
    {
        return !(*this == r);
    }
}
