#ifndef NPSTAT_LAPACK_INTERFACE_HH_
#define NPSTAT_LAPACK_INTERFACE_HH_

/*!
// \file lapack_interface.hh
//
// \brief Template interface to LAPACK (in its original F77 implementation)
//
// Typically, generic templates will throw an appropriate exception,
// while templates specialized for float, double, etc. will perform
// the requested calculations via corresponding LAPACK routines.
//
// This code is rather low-level. It may be more convenient to use
// the API provided by the Matrix class which will eventually call
// appropriate LAPACK interface functions from here.
//
// Author: I. Volobouev
//
// September 2010
*/

namespace npstat {
    /**
    // Invert a positive definite symmetric matrix
    // (LAPACK DPOTRF/DPOTRI routines are used for doubles)
    */
    template<typename Numeric>
    void invert_posdef_sym_matrix(const Numeric* in, unsigned dim,
                                  Numeric* out);

    /**
    // Invert a symmetric matrix
    // (LAPACK DSYTRF/DSYTRI routines are used for doubles)
    */
    template<typename Numeric>
    void invert_sym_matrix(const Numeric* in, unsigned dim, Numeric* out);

    /**
    // Invert a general matrix
    // (LAPACK DGETRF/DGETRI routines are used for doubles)
    */
    template<typename Numeric>
    void invert_general_matrix(const Numeric* in, unsigned dim, Numeric* out);

    /**
    // Determine eigenvalues of a symmetric matrix
    // (LAPACK DSYEV routine is used for doubles)
    */
    template<typename Numeric>
    void sym_matrix_eigenvalues(const Numeric* in, unsigned dim,
                                Numeric* eigenvalues);

    /**
    // Determine eigenvalues of a symmetric matrix using the divide
    // and conquer LAPACK driver (DSYEVD routine is used for doubles)
    */
    template<typename Numeric>
    void sym_matrix_eigenvalues_dc(const Numeric* in, unsigned dim,
                                   Numeric* eigenvalues);

    /**
    // Determine eigenvalues of a symmetric matrix using the
    // "Relatively Robust Representations" (RRR) LAPACK driver
    // (DSYEVR routine is used for doubles)
    */
    template<typename Numeric>
    void sym_matrix_eigenvalues_rrr(const Numeric* in, unsigned dim,
                                    Numeric* eigenvalues);

    /**
    // Determine eigenvalues and eigenvectors of a symmetric matrix
    // (LAPACK DSYEV routine is used for doubles)
    */
    template<typename Numeric>
    void sym_matrix_eigensystem(const Numeric* in, unsigned dim,
                                Numeric* eigenvalues,
                                Numeric* eigenvectors);

    /**
    // Determine eigenvalues and eigenvectors of a symmetric matrix
    // using the divide and conquer LAPACK driver (DSYEVD routine
    // is used for doubles)
    */
    template<typename Numeric>
    void sym_matrix_eigensystem_dc(const Numeric* in, unsigned dim,
                                   Numeric* eigenvalues,
                                   Numeric* eigenvectors);

    /**
    // Determine eigenvalues and eigenvectors of a symmetric matrix
    // using the "Relatively Robust Representations" (RRR) LAPACK
    // driver (DSYEVR routine is used for doubles)
    */
    template<typename Numeric>
    void sym_matrix_eigensystem_rrr(const Numeric* in, unsigned dim,
                                    Numeric* eigenvalues,
                                    Numeric* eigenvectors);

    /**
    // Solve a linear system (LAPACK DGETRF/DGETRS routines are used
    // for doubles). This function returns "true" on success or "false"
    // in case the matrix is degenerate.
    */
    template<typename Numeric>
    bool solve_linear_system(const Numeric* in, unsigned dim,
                             const Numeric* rhs, Numeric* solution);
}

#include "npstat/nm/lapack_interface.icc"

#endif // NPSTAT_LAPACK_INTERFACE_HH_
