#ifndef NPSTAT_ABSCOPULASMOOTHER_HH_
#define NPSTAT_ABSCOPULASMOOTHER_HH_

/*!
// \file AbsCopulaSmoother.hh
//
// \brief Interface definition for classes which build discrete copulas
//
// Author: I. Volobouev
//
// September 2010
*/

#include "npstat/stat/AbsBandwidthCV.hh"
#include "npstat/stat/empiricalCopulaHisto.hh"

namespace npstat{
    /**
    // Interface definition for classes which build discrete copulas out of data
    // samples (typically, by LOrPE, KDE, or some other nonparametric method).
    // The derived classes are supposed to override the "smoothHistogram"
    // member function.
    */
    class AbsCopulaSmoother
    {
    public:
        /**
        // Parameters "tolerance" and "maxNormalizationCycles" tells us
        // how many copula normalization cycles to perform after smoothing.
        // These cycles make sure that the resulting distribution has
        // uniform marginal densities. The cycles are needed because,
        // in general, marginals are not preserved by smoothing
        // procedures.
        */
        AbsCopulaSmoother(const unsigned* nBinsInEachDim, unsigned dim,
                          double tolerance, unsigned maxNormalizationCycles);
        virtual ~AbsCopulaSmoother() {}

        /** The "bandwidthUsed" pointer in the function below can be NULL */
        template <class Point>
        const HistoND<double>& smooth(
            std::vector<OrderedPointND<Point> >& in, double* bandwidthUsed);

        /** Return the shape of the copula array */
        inline ArrayShape copulaShape() const 
            {return h_.binContents().shape();}

    private:
        AbsCopulaSmoother();
        AbsCopulaSmoother(const AbsCopulaSmoother&);
        AbsCopulaSmoother& operator=(const AbsCopulaSmoother&);

        // The "bandwidthUsed" pointer in the function below can be NULL
        virtual void smoothHistogram(HistoND<double>& histo,
                                     double* bandwidthUsed) = 0;
        HistoND<double> h_;
        double tol_;
        unsigned nCycles_;
    };
}

#include "npstat/stat/AbsCopulaSmoother.icc"

#endif // NPSTAT_ABSCOPULASMOOTHER_HH_
