#include "geners/binaryIO.hh"
#include "npstat/stat/Distribution1DReader.hh"
#include "npstat/nm/GaussLegendreQuadrature.hh"

namespace npstat {
    bool AbsScalableDistribution1D::write(std::ostream& os) const
    {
        gs::write_pod(os, location_);
        gs::write_pod(os, scale_);
        return !os.fail();
    }

    bool AbsScalableDistribution1D::read(std::istream& is,
                                         double* location, double* scale)
    {
        gs::read_pod(is, location);
        gs::read_pod(is, scale);
        return !is.fail();
    }

    AbsDistribution1D* AbsDistribution1D::read(const gs::ClassId& id,
                                               std::istream& in)
    {
        return StaticDistribution1DReader::instance().read(id, in);
    }

    DensityScan1D::DensityScan1D(const AbsDistribution1D& fcn, 
                                 const double normfactor,
                                 const unsigned long nbins,
                                 const double xmin, const double xmax,
                                 const unsigned nIntegrationPoints)
        : fcn_(fcn),
          norm_(normfactor),
          xmin_(xmin),
          bw_((xmax-xmin)/nbins),
          nPoints_(nIntegrationPoints)
    {
        if (!nbins) throw std::invalid_argument(
            "In npstat::DensityScan1D constructor: "
            "number of bins must be positive");
        if (!nPoints_) throw std::invalid_argument(
            "In npstat::DensityScan1D constructor: "
            "number of integration points per bin must be positive");
        if (nPoints_ > 1U)
        {
            GaussLegendreQuadrature q(nPoints_);
            if (!q.npoints()) throw std::invalid_argument(
                "In npstat::DensityScan1D constructor: this number "
                "of integration points per bin is not supported by "
                "npstat::GaussLegendreQuadrature");
            weights_.resize(nPoints_/2U);
            a_.resize(nPoints_/2U);
            q.getWeights(&weights_[0], weights_.size());
            q.getAbscissae(&a_[0], a_.size());
        }
    }

    double DensityScan1D::operator()(const unsigned* index, unsigned len) const
    {
        if (len != 1U) throw std::invalid_argument(
            "In npstat::DensityScan1D::operator(): "
            "incompatible dimensionality of the input index");
        assert(index);
        const double midpoint = xmin_ + (*index + 0.5)*bw_;
        if (nPoints_ == 1U)
            return fcn_.density(midpoint)*norm_;
        else
        {
            const long double *w = &weights_[0];
            const long double *ab = &a_[0];
            const unsigned halfpoints = nPoints_/2U;
            long double sum = 0.0L;
            for (unsigned i=0; i<halfpoints; ++i)
            {
                const double delta = bw_/2.0*static_cast<double>(ab[i]);
                sum += w[i]*fcn_.density(midpoint - delta);
                sum += w[i]*fcn_.density(midpoint + delta);
            }
            return norm_*static_cast<double>(sum/2.0L);
        }
    }
}
