#ifndef NPSTAT_ABSDISTRIBUTION1D_HH_
#define NPSTAT_ABSDISTRIBUTION1D_HH_

/*!
// \file AbsDistribution1D.hh
//
// \brief Interface definition for 1-d continuous statistical distributions
//
// Author: I. Volobouev
//
// November 2009
*/

#include <cassert>
#include <typeinfo>
#include <stdexcept>

#include "geners/ClassId.hh"
#include "npstat/nm/SimpleFunctors.hh"
#include "npstat/rng/AbsRandomGenerator.hh"

namespace npstat {
    /**
    // All classes derived from this base should have copy constructors
    // (possibly automatic).
    //
    // quantile(0.0) and quantile(1.0) should return the boundaries
    // of the support interval.
    */
    struct AbsDistribution1D
    {
        inline virtual ~AbsDistribution1D() {}

        /** Probability density */
        virtual double density(double x) const = 0;

        /** Cumulative distribution function */
        virtual double cdf(double x) const = 0;

        /** 1 - cdf, implementations should avoid subtractive cancellation */
        virtual double exceedance(double x) const = 0;

        /** The quantile function */
        virtual double quantile(double x) const = 0;

        /**
        // Derived classes should not implement "operator==", implement
        // "isEqual" instead
        */
        inline bool operator==(const AbsDistribution1D& r) const
            {return (typeid(*this) == typeid(r)) && this->isEqual(r);}

        /** Logical negation of operator== */
        inline bool operator!=(const AbsDistribution1D& r) const
            {return !(*this == r);}

        /** "Virtual copy constructor" */
        virtual AbsDistribution1D* clone() const = 0;

        /**
        // Random number generator according to the given distribution.
        // Should return the number of random points used up from the
        // generator.
        */
        inline virtual unsigned random(AbsRandomGenerator& g,
                                       double* generatedRandom) const
        {
            assert(generatedRandom);
            *generatedRandom = quantile(g());
            return 1U;
        }

        //@{
        /** Prototype needed for I/O */
        virtual gs::ClassId classId() const = 0;
        virtual bool write(std::ostream&) const {return false;}
        //@}

        static inline const char* classname()
            {return "npstat::AbsDistribution1D";}
        static inline unsigned version() {return 1;}
        static AbsDistribution1D* read(const gs::ClassId& id, std::istream&);

    protected:
        /** Comparison for equality. To be implemented by derived classes. */
        virtual bool isEqual(const AbsDistribution1D&) const = 0;

#ifdef SWIG
    public:
        inline double random2(AbsRandomGenerator& g) const
        {
            double d;
            this->random(g, &d);
            return d;
        }
#endif
    };

    /**
    // This base class is used to model distributions which have
    // "trivial" behavior w.r.t. location and scale parameters. That is,
    // if the distribution density is g(x), the density with location
    // parameter "mu" and scale parameter "sigma" is g((x - mu)/sigma)/sigma.
    //
    // All distributions with infinite support on both sides fall in this
    // category, as well as a large number of distributions with finite
    // support.
    */
    class AbsScalableDistribution1D : public AbsDistribution1D
    {
    public:
        /** Location and scale parameters must be provided */
        inline AbsScalableDistribution1D(const double location,
                                         const double scale) :
            location_(location), scale_(scale)
        {
            if (scale_ <= 0.0) throw std::invalid_argument(
                "In npstat::AbsScalableDistribution1D constructor: "
                "scale parameter must be positive");
        }

        inline virtual ~AbsScalableDistribution1D() {}

        /** Get the location parameter */
        inline double location() const {return location_;}

        /** Get the scale parameter */
        inline double scale() const {return scale_;}

        /** Set the location parameter */
        inline void setLocation(const double v) {location_ = v;}

        /** Set the scale parameter */
        inline void setScale(const double v)
        {
            if (v <= 0.0) throw std::invalid_argument(
                "In npstat::AbsScalableDistribution1D::setScale: "
                "scale parameter must be positive");
            scale_ = v;
        }

        //@{
        /** Method overriden from the AbsDistribution1D base class */
        inline double density(const double x) const
            {return unscaledDensity((x - location_)/scale_)/scale_;}

        inline double cdf(const double x) const
            {return unscaledCdf((x - location_)/scale_);}

        inline double exceedance(const double x) const
            {return unscaledExceedance((x - location_)/scale_);}

        inline double quantile(const double x) const
            {return scale_*unscaledQuantile(x) + location_;}
        //@}

        /** "Virtual copy constructor" */
        virtual AbsScalableDistribution1D* clone() const = 0;

        //@{
        /** Method related to "geners" I/O */
        virtual gs::ClassId classId() const = 0;
        virtual bool write(std::ostream& os) const;
        //@}

        /**
        // Pseudo-read function for I/O. Note that this class is abstract,
        // so its instance can not be created.
        */
        static bool read(std::istream& is, double* location, double* scale);

    protected:
        /**
        // Derived classes should override the following method as long as
        // they have at least one additional data member. Don't forget to
        // call "isEqual" of the base class inside the derived classes.
        */
        virtual bool isEqual(const AbsDistribution1D& other) const
        {
            const AbsScalableDistribution1D& r = 
                static_cast<const AbsScalableDistribution1D&>(other);
            return location_ == r.location_ && scale_ == r.scale_;
        }

    private:
        AbsScalableDistribution1D();

        virtual double unscaledDensity(double x) const = 0;
        virtual double unscaledCdf(double x) const = 0;
        virtual double unscaledExceedance(double x) const = 0;
        virtual double unscaledQuantile(double x) const = 0;

        double location_;
        double scale_;
    };

    /** A functor for the density function of the given 1-d distribution */
    class DensityFunctor1D : public Functor1<double, double>
    {
    public:
        inline DensityFunctor1D(const AbsDistribution1D& fcn,
                                const double normfactor=1.0)
            : fcn_(fcn), norm_(normfactor) {}

        inline double operator()(const double& a) const
            {return norm_*fcn_.density(a);}
    private:
        DensityFunctor1D();
        const AbsDistribution1D& fcn_;
        const double norm_;
    };

    /** A functor for filling one-dimensional ArrayND with density values */
    class DensityScan1D
    {
    public:
        /**
        // Parameter "nIntegrationPoints" determines how many points per bin
        // will be used to calculate the bin average (Gauss-Legendre quadrature
        // used). This number must be either 1 or one of the numbers of points
        // supported by the "GaussLegendreQuadrature" class.
        */
        DensityScan1D(const AbsDistribution1D& fcn, double normfactor,
                      unsigned long nbins, double xmin, double xmax,
                      unsigned nIntegrationPoints=1);

        double operator()(const unsigned* index, unsigned len) const;

    private:
        DensityScan1D();

        const AbsDistribution1D& fcn_;
        double norm_;
        double xmin_;
        double bw_;
        std::vector<long double> a_;
        std::vector<long double> weights_;
        unsigned nPoints_;
    };

    /**
    // A helper functor to be used for the determination of density
    // discretization errors (when the density is represented by
    // a collection of values on a grid)
    */
    class DensityDiscretizationError1D : public Functor1<double, double>
    {
    public:
        /** 
        // "discreteValue" is the density value on a particular
        // discretization interval. It is expected that the driver
        // code will integrate the L2 error returned by this functor
        // over all such intervals.
        */
        inline DensityDiscretizationError1D(const AbsDistribution1D& fcn,
                                            const double normfactor,
                                            const double discreteValue)
            : fcn_(fcn), norm_(normfactor), h_(discreteValue) {}

        inline double operator()(const double& a) const
        {
            const double d = norm_*fcn_.density(a) - h_;
            return d*d;
        }

    private:
        DensityDiscretizationError1D();
        const AbsDistribution1D& fcn_;
        const double norm_;
        const double h_;
    };

    /**
    // A functor for filling one-dimensional ArrayND with density values.
    //
    // Similar to DensityScan1D but can be used with an arbitrary transform
    // of the array index. The arbitrariness of the transform also prevents
    // us from integrating over the bin size.
    */
    template<class Transform>
    class DensityScan1DTrans
    {
    public:
        /**
        // The "transform" parameter will not be copied intrernally. The user
        // is responsible for ensuring the proper lifetime of this object.
        */
        inline DensityScan1DTrans(const AbsDistribution1D& fcn,
                                  const Transform& transform,
                                  const double normfactor) :
            fcn_(fcn), mapping_(transform), norm_(normfactor) {}

        inline double operator()(const unsigned* index, const unsigned len) const
        {
            if (len != 1U) throw std::invalid_argument(
                "In npstat::DensityScan1DTrans::operator(): "
                "incompatible input point dimensionality");
            assert(index);
            const double x = mapping_(static_cast<double>(index[0]));
            return norm_*fcn_.density(x);
        }

    private:
        DensityScan1DTrans();

        const AbsDistribution1D& fcn_;
        const Transform& mapping_;
        double norm_;
    };
}

#endif // NPSTAT_ABSDISTRIBUTION1D_HH_
