#ifndef NPSTAT_DISTRIBUTIONS1D_HH_
#define NPSTAT_DISTRIBUTIONS1D_HH_

/*!
// \file Distributions1D.hh
//
// \brief A number of useful 1-d continuous statistical distributions
//
// Author: I. Volobouev
//
// November 2009
*/

#include "npstat/stat/Distribution1DFactory.hh"

namespace npstat {
    /**
    // The uniform distribution is defined here by a constant density
    // equal to 1 between 0 and 1 and equal to 0 everywhere else
    */
    class Uniform1D : public AbsScalableDistribution1D
    {
    public:
        inline Uniform1D(const double location, const double scale)
            : AbsScalableDistribution1D(location, scale) {}
        inline Uniform1D* clone() const {return new Uniform1D(*this);}

        // Methods needed for I/O
        virtual gs::ClassId classId() const {return gs::ClassId(*this);}

        static inline const char* classname() {return "npstat::Uniform1D";}
        static inline unsigned version() {return 1;}
        static Uniform1D* read(const gs::ClassId& id, std::istream& in);

    protected:
        virtual bool isEqual(const AbsDistribution1D& r) const
            {return AbsScalableDistribution1D::isEqual(r);}

    private:
        friend class ScalableDistribution1DFactory<Uniform1D>;

        inline Uniform1D(const double location, const double scale,
                         const std::vector<double>& /* params */)
            : AbsScalableDistribution1D(location, scale) {}
        inline static int nParameters() {return 0;}

        double unscaledDensity(double x) const;
        double unscaledCdf(double x) const;
        double unscaledQuantile(double x) const;

        inline double unscaledExceedance(const double x) const
            {return 1.0 - unscaledCdf(x);}
    };

    /** Exponential distribution. "scale" is the decay time. */
    class Exponential1D : public AbsScalableDistribution1D
    {
    public:
        inline Exponential1D(const double location, const double scale)
            : AbsScalableDistribution1D(location, scale) {}
        inline Exponential1D* clone() const {return new Exponential1D(*this);}

        // Methods needed for I/O
        virtual gs::ClassId classId() const {return gs::ClassId(*this);}

        static inline const char* classname() {return "npstat::Exponential1D";}
        static inline unsigned version() {return 1;}
        static Exponential1D* read(const gs::ClassId& id, std::istream& in);

    protected:
        virtual bool isEqual(const AbsDistribution1D& r) const
            {return AbsScalableDistribution1D::isEqual(r);}

    private:
        friend class ScalableDistribution1DFactory<Exponential1D>;
        
        inline Exponential1D(const double location, const double scale,
                             const std::vector<double>& /* params */)
            : AbsScalableDistribution1D(location, scale) {}
        inline static int nParameters() {return 0;}

        double unscaledDensity(double x) const;
        double unscaledCdf(double x) const;
        double unscaledQuantile(double x) const;
        double unscaledExceedance(double x) const;
    };

    /**
    // A distribution whose density has a simple quadratic shape.
    // The support is from 0 to 1, and the coefficients "a" and "b"
    // are the coefficients for the Legendre polynomials of 1st
    // and 2nd degree translated to the support region. Note that
    // only those values of "a" and "b" that guarantee non-negativity
    // of the density are allowed, otherwise the code will generate
    // a run-time error.
    */
    class Quadratic1D : public AbsScalableDistribution1D
    {
    public:
        Quadratic1D(double location, double scale, double a, double b);
        inline Quadratic1D* clone() const {return new Quadratic1D(*this);}

        inline double a() const {return a_;}
        inline double b() const {return b_;}

        // Methods needed for I/O
        virtual gs::ClassId classId() const {return gs::ClassId(*this);}
        virtual bool write(std::ostream& os) const;

        static inline const char* classname() {return "npstat::Quadratic1D";}
        static inline unsigned version() {return 1;}
        static Quadratic1D* read(const gs::ClassId& id, std::istream& in);

    protected:
        virtual bool isEqual(const AbsDistribution1D&) const;

    private:
        friend class ScalableDistribution1DFactory<Quadratic1D>;
        
        Quadratic1D(double location, double scale,
                    const std::vector<double>& params);
        inline static int nParameters() {return 2;}

        void verifyNonNegative();
        double unscaledDensity(double x) const;
        double unscaledCdf(double x) const;
        double unscaledQuantile(double x) const;

        inline double unscaledExceedance(const double x) const
            {return 1.0 - unscaledCdf(x);}

        double a_;
        double b_;
    };

    /**
    // A distribution whose density logarithm has a simple quadratic
    // shape. The support is from 0 to 1, and the coefficients "a" and "b"
    // are the coefficients for the Legendre polynomials of 1st and 2nd
    // degree translated to the support region.
    */
    class LogQuadratic1D : public AbsScalableDistribution1D
    {
    public:
        LogQuadratic1D(double location, double scale, double a, double b);
        inline LogQuadratic1D* clone() const
            {return new LogQuadratic1D(*this);}

        inline double a() const {return a_;}
        inline double b() const {return b_;}

        // Methods needed for I/O
        virtual gs::ClassId classId() const {return gs::ClassId(*this);}
        virtual bool write(std::ostream& os) const;

        static inline const char* classname() {return "npstat::LogQuadratic1D";}
        static inline unsigned version() {return 1;}
        static LogQuadratic1D* read(const gs::ClassId& id, std::istream& in);

    protected:
        virtual bool isEqual(const AbsDistribution1D&) const;

    private:
        friend class ScalableDistribution1DFactory<LogQuadratic1D>;

        LogQuadratic1D(double location, double scale,
                       const std::vector<double>& params);
        inline static int nParameters() {return 2;}

        void normalize();
        long double quadInteg(long double x) const;
        double unscaledDensity(double x) const;
        double unscaledCdf(double x) const;
        double unscaledQuantile(double x) const;

        inline double unscaledExceedance(const double x) const
            {return 1.0 - unscaledCdf(x);}

        long double ref_;
        long double range_;
        double a_;
        double b_;
        double k_;
        double s_;
        double norm_;
    };

    /** The Gaussian (or Normal) distribution */
    class Gauss1D : public AbsScalableDistribution1D
    {
    public:
        inline Gauss1D(const double location, const double scale)
            : AbsScalableDistribution1D(location, scale) {}
        inline Gauss1D* clone() const {return new Gauss1D(*this);}

        // Higher quality generator than the one provided by
        // the quantile function
        virtual unsigned random(AbsRandomGenerator& g,
                                double* generatedRandom) const;

        // Methods needed for I/O
        virtual gs::ClassId classId() const {return gs::ClassId(*this);}

        static inline const char* classname() {return "npstat::Gauss1D";}
        static inline unsigned version() {return 1;}
        static Gauss1D* read(const gs::ClassId& id, std::istream& in);

    protected:
        virtual bool isEqual(const AbsDistribution1D& r) const
            {return AbsScalableDistribution1D::isEqual(r);}

    private:
        friend class ScalableDistribution1DFactory<Gauss1D>;

        inline Gauss1D(const double location, const double scale,
                       const std::vector<double>& /* params */)
            : AbsScalableDistribution1D(location, scale) {}
        inline static int nParameters() {return 0;}

        double unscaledDensity(double x) const;
        double unscaledCdf(double x) const;
        double unscaledQuantile(double x) const;
        double unscaledExceedance(double x) const;
    };

    /** Gaussian distribution truncated at some number of sigmas */
    class TruncatedGauss1D : public AbsScalableDistribution1D
    {
    public:
        TruncatedGauss1D(double location, double scale, double nsigma);
        inline TruncatedGauss1D* clone() const
            {return new TruncatedGauss1D(*this);}

        inline double nsigma() const {return nsigma_;}

        // Higher quality generator than the one provided by
        // the quantile function
        virtual unsigned random(AbsRandomGenerator& g,
                                double* generatedRandom) const;

        // Methods needed for I/O
        virtual gs::ClassId classId() const {return gs::ClassId(*this);}
        virtual bool write(std::ostream& os) const;

        static inline const char* classname() {return "npstat::TruncatedGauss1D";}
        static inline unsigned version() {return 1;}
        static TruncatedGauss1D* read(const gs::ClassId& id, std::istream& in);

    protected:
        virtual bool isEqual(const AbsDistribution1D&) const;

    private:
        friend class ScalableDistribution1DFactory<TruncatedGauss1D>;

        TruncatedGauss1D(double location, double scale,
                         const std::vector<double>& params);
        inline static int nParameters() {return 1;}

        void initialize();
        double unscaledDensity(double x) const;
        double unscaledCdf(double x) const;
        double unscaledQuantile(double x) const;

        inline double unscaledExceedance(const double x) const
            {return unscaledCdf(-x);}

        double nsigma_;
        double norm_;
        double cdf0_;
    };

    /** Symmetric beta distribution */
    class SymmetricBeta1D : public AbsScalableDistribution1D
    {
    public:
        SymmetricBeta1D(double location, double scale, double power);
        inline SymmetricBeta1D* clone() const
            {return new SymmetricBeta1D(*this);}

        inline double power() const {return n_;}

        // Methods needed for I/O
        virtual gs::ClassId classId() const {return gs::ClassId(*this);}
        virtual bool write(std::ostream& os) const;

        static inline const char* classname() {return "npstat::SymmetricBeta1D";}
        static inline unsigned version() {return 1;}
        static SymmetricBeta1D* read(const gs::ClassId& id, std::istream& in);

    protected:
        virtual bool isEqual(const AbsDistribution1D&) const;

    private:
        friend class ScalableDistribution1DFactory<SymmetricBeta1D>;

        SymmetricBeta1D(double location, double scale,
                        const std::vector<double>& params);
        inline static int nParameters() {return 1;}

        double unscaledDensity(double x) const;
        double unscaledCdf(double x) const;
        double unscaledQuantile(double x) const;

        inline double unscaledExceedance(const double x) const
            {return unscaledCdf(-x);}

        double calculateNorm() const;

        double n_;
        double norm_;
    };

    /** Beta1D density is proportional to x^(apha-1) * (1-x)^(beta-1) */
    class Beta1D : public AbsScalableDistribution1D
    {
    public:
        Beta1D(double location, double scale, double alpha, double beta);
        inline Beta1D* clone() const
            {return new Beta1D(*this);}

        inline double alpha() const {return alpha_;}
        inline double beta() const {return beta_;}

        // Methods needed for I/O
        virtual gs::ClassId classId() const {return gs::ClassId(*this);}
        virtual bool write(std::ostream& os) const;

        static inline const char* classname() {return "npstat::Beta1D";}
        static inline unsigned version() {return 1;}
        static Beta1D* read(const gs::ClassId& id, std::istream& in);

    protected:
        virtual bool isEqual(const AbsDistribution1D&) const;

    private:
        friend class ScalableDistribution1DFactory<Beta1D>;

        Beta1D(double location, double scale,
               const std::vector<double>& params);
        inline static int nParameters() {return 2;}

        double unscaledDensity(double x) const;
        double unscaledCdf(double x) const;
        double unscaledExceedance(double x) const;
        double unscaledQuantile(double x) const;

        double calculateNorm() const;

        double alpha_;
        double beta_;
        double norm_;
    };

    /** "Huber" distribution */
    class Huber1D : public AbsScalableDistribution1D
    {
    public:
        Huber1D(double location, double scale, double tailWeight);
        inline Huber1D* clone() const {return new Huber1D(*this);}

        inline double tailWeight() const {return tailWeight_;}
        inline double tailStart() const {return a_;}

        // Methods needed for I/O
        virtual gs::ClassId classId() const {return gs::ClassId(*this);}
        virtual bool write(std::ostream& os) const;

        static inline const char* classname() {return "npstat::Huber1D";}
        static inline unsigned version() {return 1;}
        static Huber1D* read(const gs::ClassId& id, std::istream& in);

    protected:
        virtual bool isEqual(const AbsDistribution1D&) const;

    private:
        friend class ScalableDistribution1DFactory<Huber1D>;

        Huber1D(double location, double scale,
                const std::vector<double>& params);
        inline static int nParameters() {return 1;}

        void initialize();
        double unscaledDensity(double x) const;
        double unscaledCdf(double x) const;
        double unscaledQuantile(double x) const;

        inline double unscaledExceedance(const double x) const
            {return unscaledCdf(-x);}

        double weight(double a) const;

        double tailWeight_;
        double a_;
        double normfactor_;
        double support_;
        double cdf0_;
    };

    /** Cauchy (or Breit-Wigner) distribution */
    class Cauchy1D : public AbsScalableDistribution1D
    {
    public:
        Cauchy1D(double location, double scale);
        inline Cauchy1D* clone() const {return new Cauchy1D(*this);}

        // Methods needed for I/O
        virtual gs::ClassId classId() const {return gs::ClassId(*this);}

        static inline const char* classname() {return "npstat::Cauchy1D";}
        static inline unsigned version() {return 1;}
        static Cauchy1D* read(const gs::ClassId& id, std::istream& in);

    protected:
        virtual bool isEqual(const AbsDistribution1D& r) const
            {return AbsScalableDistribution1D::isEqual(r);}

    private:
        friend class ScalableDistribution1DFactory<Cauchy1D>;

        inline Cauchy1D(const double location, const double scale,
                        const std::vector<double>& /* params */)
            : AbsScalableDistribution1D(location, scale) {}
        inline static int nParameters() {return 0;}

        double unscaledDensity(double x) const;
        double unscaledCdf(double x) const;
        double unscaledQuantile(double x) const;

        inline double unscaledExceedance(const double x) const
            {return unscaledCdf(-x);}

        double support_;
    };

    /**
    // Log-normal distribution represented by its mean, standard
    // deviation, and skewness. This representation is more useful
    // than other representations encountered in statistical literature.
    */
    class LogNormal : public AbsScalableDistribution1D
    {
    public:
        LogNormal(double mean, double stdev, double skewness);
        inline LogNormal* clone() const {return new LogNormal(*this);}
        inline double skewness() const {return skew_;}

        // Methods needed for I/O
        virtual gs::ClassId classId() const {return gs::ClassId(*this);}
        virtual bool write(std::ostream& os) const;

        static inline const char* classname() {return "npstat::LogNormal";}
        static inline unsigned version() {return 1;}
        static LogNormal* read(const gs::ClassId& id, std::istream& in);

    protected:
        virtual bool isEqual(const AbsDistribution1D&) const;

    private:
        friend class ScalableDistribution1DFactory<LogNormal>;

        LogNormal(double location, double scale,
                  const std::vector<double>& params);
        inline static int nParameters() {return 1;}

        void initialize();
        double unscaledDensity(double x) const;
        double unscaledCdf(double x) const;
        double unscaledExceedance(double x) const;
        double unscaledQuantile(double x) const;

        double skew_;

        double logw_;
        double s_;
        double xi_;
        double emgamovd_;
    };

    /** Student's t-distribution */
    class StudentsT1D : public AbsScalableDistribution1D
    {
    public:
        StudentsT1D(double location, double scale, double nDegreesOfFreedom);
        inline StudentsT1D* clone() const {return new StudentsT1D(*this);}

        inline double nDegreesOfFreedom() const {return nDoF_;}

        // Methods needed for I/O
        virtual gs::ClassId classId() const {return gs::ClassId(*this);}
        virtual bool write(std::ostream& os) const;

        static inline const char* classname() {return "npstat::StudentsT1D";}
        static inline unsigned version() {return 1;}
        static StudentsT1D* read(const gs::ClassId& id, std::istream& in);

    protected:
        virtual bool isEqual(const AbsDistribution1D&) const;

    private:
        friend class ScalableDistribution1DFactory<StudentsT1D>;

        StudentsT1D(double location, double scale,
                    const std::vector<double>& params);
        inline static int nParameters() {return 1;}

        void initialize();
        double effectiveSupport() const;
        double unscaledDensity(double x) const;
        double unscaledCdf(double x) const;
        double unscaledExceedance(double x) const;
        double unscaledQuantile(double x) const;

        double nDoF_;
        double normfactor_;
        double power_;
        double bignum_;
    };

    /** Distribution defined by an interpolation table */
    class Tabulated1D : public AbsScalableDistribution1D
    {
    public:
        // The "data" array gives (unnormalized) density values at
        // equidistant intervals. data[0] is density at 0.0, and
        // data[dataLen-1] is density at 1.0. If "dataLen" is less
        // than 2, uniform distribution will be created. Internally,
        // the data is kept in double precision.
        //
        // "interpolationDegree" must be less than 4 and less than "dataLen".
        //
        template <typename Real>
        Tabulated1D(double location, double scale,
                    const Real* data, unsigned dataLen,
                    unsigned interpolationDegree);

        inline Tabulated1D(const double location, const double scale,
                           const std::vector<double>& table,
                           const unsigned interpolationDegree)
            : AbsScalableDistribution1D(location, scale)
        {
            const unsigned long sz = table.size();
            initialize(sz ? &table[0] : (double*)0, sz, interpolationDegree);
        }

        inline virtual Tabulated1D* clone() const
            {return new Tabulated1D(*this);}

        inline virtual ~Tabulated1D() {}

        inline unsigned interpolationDegree() const {return deg_;}
        inline unsigned tableLength() const {return len_;}
        inline const double* tableData() const {return &table_[0];}

        // Methods needed for I/O
        virtual gs::ClassId classId() const {return gs::ClassId(*this);}
        virtual bool write(std::ostream& os) const;

        static inline const char* classname() {return "npstat::Tabulated1D";}
        static inline unsigned version() {return 1;}
        static Tabulated1D* read(const gs::ClassId& id, std::istream& in);

    protected:
        virtual bool isEqual(const AbsDistribution1D&) const;

    private:
        friend class ScalableDistribution1DFactory<Tabulated1D>;

        // The following constructor creates interpolator
        // of maximum degree possible
        Tabulated1D(double location, double scale,
                    const std::vector<double>& params);
        inline static int nParameters() {return -1;}

        double unscaledDensity(double x) const;
        double unscaledCdf(double x) const;
        double unscaledQuantile(double x) const;

        inline double unscaledExceedance(const double x) const
            {return 1.0 - unscaledCdf(x);}

        template <typename Real> void initialize(
            const Real* data, unsigned dataLen, unsigned interpolationDegree);

        void normalize();
        double interpolate(double x) const;
        double intervalInteg(unsigned intervalNumber) const;
        double interpIntegral(double x0, double x1) const;

        std::vector<double> table_;
        std::vector<double> cdf_;
        double step_;
        unsigned len_;
        unsigned deg_;
    };

    /**
    // Another interpolated distribution. For this one, we will assume
    // that the coordinates correspond to 1-d histogram bin centers.
    */
    class BinnedDensity1D : public AbsScalableDistribution1D
    {
    public:
        // The "data" array gives density values at equidistant intervals.
        // data[0] is density at 0.5/dataLen, and data[dataLen-1] is density
        // at 1.0 - 0.5/dataLen.
        //
        // "interpolationDegree" must be less than 2.
        //
        template <typename Real>
        BinnedDensity1D(double location, double scale,
                        const Real* data, unsigned dataLen,
                        unsigned interpolationDegree);

        inline BinnedDensity1D(const double location, const double scale,
                               const std::vector<double>& table,
                               const unsigned interpolationDegree)
            : AbsScalableDistribution1D(location, scale)
        {
            const unsigned long sz = table.size();
            initialize(sz ? &table[0] : (double*)0, sz, interpolationDegree);
        }

        inline virtual BinnedDensity1D* clone() const
            {return new BinnedDensity1D(*this);}

        inline virtual ~BinnedDensity1D() {}

        inline unsigned interpolationDegree() const {return deg_;}
        inline unsigned tableLength() const {return len_;}
        inline const double* tableData() const {return &table_[0];}

        // Methods needed for I/O
        virtual gs::ClassId classId() const {return gs::ClassId(*this);}
        virtual bool write(std::ostream& os) const;

        static inline const char* classname() {return "npstat::BinnedDensity1D";}
        static inline unsigned version() {return 1;}
        static BinnedDensity1D* read(const gs::ClassId& id, std::istream& in);

    protected:
        virtual bool isEqual(const AbsDistribution1D&) const;

    private:
        friend class ScalableDistribution1DFactory<BinnedDensity1D>;

        // The following constructor creates interpolator
        // of maximum degree possible
        BinnedDensity1D(double location, double scale,
                        const std::vector<double>& params);
        inline static int nParameters() {return -1;}

        double unscaledDensity(double x) const;
        double unscaledCdf(double x) const;
        double unscaledQuantile(double x) const;

        inline double unscaledExceedance(const double x) const
            {return 1.0 - unscaledCdf(x);}

        template <typename Real> void initialize(
            const Real* data, unsigned dataLen, unsigned interpolationDegree);

        void normalize();
        double interpolate(double x) const;

        std::vector<double> table_;
        std::vector<double> cdf_;
        double step_;
        unsigned len_;
        unsigned deg_;
        unsigned firstNonZeroBin_;
        unsigned lastNonZeroBin_;
    };
}

#include "npstat/stat/Distributions1D.icc"

#endif // NPSTAT_DISTRIBUTIONS1D_HH_
