#include <cmath>
#include <cassert>
#include <stdexcept>
#include <climits>

#include "npstat/nm/OrthoPoly1D.hh"
#include "npstat/stat/Filter1DBuilders.hh"

static npstat::PolyFilter1D* build(
    const double* taper, const unsigned maxDegree,
    const double* weight, const unsigned len,
    const unsigned center, const double step)
{
    npstat::OrthoPoly1D poly(maxDegree, weight, len, step);
    npstat::PolyFilter1D* f = new npstat::PolyFilter1D(center);
    f->resize(len);
    poly.linearFilter(taper, maxDegree, center, &(*f)[0], len);
    return f;
}

namespace npstat {
    WeightTableFilter1DBuilder::WeightTableFilter1DBuilder(
        const double* weight, const unsigned weightLen)
    {
        if (!weightLen) throw std::invalid_argument(
            "In npstat::WeightTableFilter1DBuilder constructor: "
            "array of weights must not be empty");
        assert(weight);

        // Construct the full weight
        w_.resize(2*weightLen - 1);
        w_[weightLen - 1] = weight[0];       
        for (unsigned i=1; i<weightLen; ++i)
        {
            w_[weightLen + i - 1] = weight[i];
            w_[weightLen - i - 1] = weight[i];
        }
    }

    PolyFilter1D* WeightTableFilter1DBuilder::makeFilter(
        const double* taper, const unsigned maxDeg,
        const unsigned i, const unsigned dataLen) const
    {
        if (!dataLen) throw std::invalid_argument(
            "In npstat::WeightTableFilter1DBuilder::makeFilter: "
            "invalid data length parameter");
        assert(taper);

        const double step = 1.0/sqrt(1.0*w_.size());
        const unsigned n_half = w_.size()/2;
        if (i < n_half)
        {
            // Create left boundary weight and polynomials
            unsigned n_use = n_half + i + 1;
            if (n_use > dataLen)
                n_use = dataLen;
            return build(taper, maxDeg, &w_[n_half-i], n_use, i, step);
        }
        else if (i + n_half >= dataLen)
        {
            // Create right boundary weight and polynomials
            const unsigned n_use = dataLen + n_half - i;
            assert(n_use <= dataLen);
            return build(taper, maxDeg, &w_[0], n_use, n_half, step);
        }
        else
        {
            // Symmetric weight
            return build(taper, maxDeg, &w_[0], w_.size(), n_half, step);
        }
    }

    StretchingFilter1DBuilder::StretchingFilter1DBuilder(
        const AbsDistribution1D* distro, const double step)
        : distro_(distro),
          step_(step)
    {
        if (step_ <= 0.0) throw std::invalid_argument(
            "In npstat::StretchingFilter1DBuilder constructor: "
            "invalid step size");
        assert(distro_);

        const double maxratio = floor(fabs(distro_->quantile(1.0))/step_);
        const double minratio = floor(fabs(distro_->quantile(0.0))/step_);
        const double topratio = maxratio > minratio ? maxratio : minratio;
        if (topratio >= static_cast<double>(INT_MAX))
            maxlen_ = 2*(static_cast<unsigned>(INT_MAX) - 1) + 1;
        else
            maxlen_ = 2*static_cast<unsigned>(topratio) + 1;
    }

    PolyFilter1D* StretchingFilter1DBuilder::makeFilter(
        const double* taper, const unsigned maxDeg,
        const unsigned i, const unsigned dataLen) const
    {
        if (!(dataLen && dataLen <= static_cast<unsigned>(INT_MAX)))
            throw std::invalid_argument(
                "In npstat::StretchingFilter1DBuilder::makeFilter: "
                "invalid data length parameter");
        assert(taper);

        // Make sure we have enough space
        const unsigned sizeNeeded = (dataLen < maxlen_ ? dataLen : maxlen_);
        if (w_.size() < sizeNeeded)
            w_.resize(sizeNeeded);

        // Figure out the maximum possible efficiency of the weight scan
        const double maxeff = distro_->cdf((dataLen - 1 - dataLen/2)*step_) - 
                              distro_->cdf(-step_*(dataLen/2));
        if (maxeff <= 0.0) throw std::invalid_argument(
            "In npstat::StretchingFilter1DBuilder::makeFilter: "
            "weight density is identically zero in the relevant region");

        // Figure out the efficiency for this bin
        const double eff = distro_->cdf((dataLen - 1 - i)*step_) - 
                           distro_->cdf(-step_*i);

        // Stretch the function so that it covers wider range if
        // the efficiency is small. Stretching the function is
        // equivalent to reducing the step size used for scanning
        // the function.
        const double scanStep = step_*eff/maxeff;

        // Figure out where the center bin is going to be
        int j0;
        const unsigned n_half = maxlen_/2;
        if (i < n_half)
        {
            // Create the left boundary weight and polynomials.
            // The center is going to be at i.
            j0 = static_cast<int>(i);
        }
        else if (i + n_half >= dataLen)
        {
            // Create right boundary weight and polynomials.
            // The center is going to be at sizeNeeded - (dataLen - i).
            assert(sizeNeeded >= dataLen - i);
            j0 = static_cast<int>(sizeNeeded - (dataLen - i));
        }
        else
        {
            // Symmetric support region
            assert(maxlen_ <= dataLen);
            j0 = static_cast<int>(n_half);
        }

        const int jmax = static_cast<int>(sizeNeeded);
        assert(j0 >= 0 && j0 < jmax);
        for (int j=0; j<jmax; ++j)
            w_[j] = distro_->density(scanStep*(j - j0));
        return build(taper, maxDeg, &w_[0], sizeNeeded, j0, step_);
    }
}
