#ifndef NPSTAT_FILTER1DBUILDERS_HH_
#define NPSTAT_FILTER1DBUILDERS_HH_

/*!
// \file Filter1DBuilders.hh
//
// \brief Build local polynomial filters in one dimension
//
// The filters are intended for use with the "LocalPolyFilter1D" class.
//
// Author: I. Volobouev
//
// November 2009
*/

#include "npstat/stat/AbsFilter1DBuilder.hh"
#include "npstat/stat/AbsDistribution1D.hh"

namespace npstat {
    /**
    // This class will construct a local polynomial filter out of
    // an arbitrary sampled even weight function and a taper function
    */
    class WeightTableFilter1DBuilder : public AbsFilter1DBuilder
    {
    public:
        /**
        // The constructor arguments are as follows:
        //
        //  weight    -- Tabulated weight function, in equidistant
        //               steps, in the order of increasing x.
        //               The first value corresponds to x = 0,
        //               the second to x + step, etc. The weight
        //               is assumed to be symmetric around 0.
        //
        //  weightLen -- The length of the "weight" array.
        */
        WeightTableFilter1DBuilder(const double* weight, unsigned weightLen);

        inline unsigned centralWeightLength() const {return w_.size();}

        PolyFilter1D* makeFilter(const double* taper, unsigned maxDegree,
                                 unsigned binnum, unsigned datalen) const;
    private:
        WeightTableFilter1DBuilder();

        std::vector<double> w_;
    };

    /**
    // This class will construct a local polynomial filter out of
    // an AbsDistribution1D weight function and a taper function
    //
    // The weight function will be stretched near the boundaries
    // by the factor Iw_max / Iw, where Iw is the integral of the
    // weight function inside the support boundary (Iw_max is the
    // same integral calculated far away from any boundary).
    */
    class StretchingFilter1DBuilder : public AbsFilter1DBuilder
    {
    public:
        /**
        // This class will not own the AbsDistribution1D object.
        // It is the responsibility of the user of this class
        // to make sure that the AbsDistribution1D object stays
        // alive while this object is in use.
        //
        // It will be assumed that the distribution has its peak at 0.0.
        */
        StretchingFilter1DBuilder(const AbsDistribution1D* distro,
                                  double centralStepSize);

        inline unsigned centralWeightLength() const {return maxlen_;}

        PolyFilter1D* makeFilter(const double* taper, unsigned maxDegree,
                                 unsigned binnum, unsigned datalen) const;
    private:
        StretchingFilter1DBuilder();

        const AbsDistribution1D* distro_;
        double step_;
        unsigned maxlen_;
        mutable std::vector<double> w_;
    };
}

#endif // NPSTAT_FILTER1DBUILDERS_HH_
