#ifndef NPSTAT_GRIDINTERPOLATEDDISTRIBUTION_HH_
#define NPSTAT_GRIDINTERPOLATEDDISTRIBUTION_HH_

/*!
// \file GridInterpolatedDistribution.hh
//
// \brief Multivariate distribution obtained by nonparametric interpolation of
//        probability densities
//
// Author: I. Volobouev
//
// July 2010
*/

#include <stdexcept>
#include <iostream>

#include "npstat/nm/GridAxis.hh"
#include "npstat/nm/ArrayND.hh"

#include "npstat/stat/AbsGridInterpolatedDistribution.hh"

namespace npstat {
    /**
    // This class interpolates between probability distributions placed at
    // the points of a rectangular (not necessarily equidistant) grid. The
    // weights in the interpolations are determined in the multilinear
    // fashion. Multivariate quantile or copula interpolation method can be
    // used, depending on the types of interpolated densities.
    //
    // The "Interpolation" template parameter should normally be one of
    // "CopulaInterpolationND" or "UnitMapInterpolationND".
    */
    template <class Interpolation>
    class GridInterpolatedDistribution : public AbsGridInterpolatedDistribution
    {
    public:
        /** 
        // Constructor arguments are the collection of parameter grid
        // axes and the dimensionality of the associated distributions
        */
        GridInterpolatedDistribution(const std::vector<GridAxis>& axes,
                                     unsigned distributionDimension);

        GridInterpolatedDistribution(const GridInterpolatedDistribution&);

        virtual ~GridInterpolatedDistribution();

        GridInterpolatedDistribution& operator=(
            const GridInterpolatedDistribution&);

        inline virtual GridInterpolatedDistribution* clone() const
            {return new GridInterpolatedDistribution(*this);}

        /**
        // Associate a distribution with the given parameter grid cell.
        //
        // This object will assume the ownership of the "distro" pointer.
        // All distributions must be set before density can be calculated
        // on all points in and outside the grid.
        */
        void setGridDistro(const unsigned* cell, unsigned lenCell,
                           AbsDistributionND* distro);

        /**
        // Associate a distribution with the given parameter grid cell
        // using a linear index for the cell
        */
        void setLinearDistro(unsigned idx, AbsDistributionND* distro);

        /**
        // The following function must be called before calculating
        // the density. The length of the coordinates array must be
        // equal to the number of grid axes.
        */
        void setGridCoords(const double* coords, unsigned lenCoords);

        //@{
        /** Inspect object properties */
        inline unsigned nAxes() const {return nAxes_;}
        inline const GridAxis& getAxis(const unsigned i) const
            {return axes_.at(i);}
        inline unsigned long nDistros() const {return interpolated_.length();}
        inline ArrayShape gridShape() const {return interpolated_.shape();}
        //@}

        inline double density(const double* x, unsigned dim) const
        {
            if (!pointSet_) throw std::logic_error(
                "In npstat::GridInterpolatedDistribution::density: "
                "grid coordinates not set");
            return interpolator_.density(x, dim);
        }

        inline void unitMap(const double* rnd, unsigned len, double* x) const
        {
            if (!pointSet_) throw std::logic_error(
                "In npstat::GridInterpolatedDistribution::unitMap: "
                "grid coordinates not set");
            return interpolator_.unitMap(rnd, len, x);
        }

        inline bool mappedByQuantiles() const
        {
            if (!pointSet_) throw std::logic_error(
                "In npstat::GridInterpolatedDistribution::mappedByQuantiles: "
                "grid coordinates not set");
            return interpolator_.mappedByQuantiles();
        }

        //@{
        // Method needed for "geners" I/O
        virtual gs::ClassId classId() const {return gs::ClassId(*this);}
        virtual bool write(std::ostream& of) const;
        //@}

        // Class name for the I/O
        static const char* classname();

        // Version number for the I/O
        static inline unsigned version() {return 1;}

        // Read method for the I/O
        static GridInterpolatedDistribution* read(
            const gs::ClassId& id, std::istream& in);

    private:
        GridInterpolatedDistribution();

        virtual bool isEqual(const AbsDistributionND& r) const;

        void releaseDistros();
        void cloneDistros();

        // We will own the interpolated distributions
        std::vector<GridAxis> axes_;
        std::vector<unsigned> cellBuf_;
        ArrayND<AbsDistributionND*> interpolated_;
        Interpolation interpolator_;
        unsigned nAxes_;
        bool pointSet_;
    };
}

#include "npstat/stat/GridInterpolatedDistribution.icc"

#endif // NPSTAT_GRIDINTERPOLATEDDISTRIBUTION_HH_
