#include <stdexcept>
#include <algorithm>

#include "npstat/nm/allocators.hh"
#include "npstat/stat/LocalPolyFilter1D.hh"
#include "npstat/stat/Distributions1D.hh"
#include "npstat/stat/Filter1DBuilders.hh"

namespace npstat {
    LocalPolyFilter1D::LocalPolyFilter1D(
        const double* itaper, const unsigned maxDegree,
        const AbsFilter1DBuilder& filterBuilder,
        const unsigned dataLen)
        : taper_(0), bins_(0), nbins_(dataLen), maxDegree_(maxDegree)
    {
        if (!dataLen) throw std::invalid_argument(
            "In npstat::LocalPolyFilter1D constructor: "
            "invalid expected data length (must be positive)");

        // Figure out the length of the weight array in the center
        const unsigned n_kernel = filterBuilder.centralWeightLength();
        if (!(n_kernel % 2)) throw std::invalid_argument(
            "In npstat::LocalPolyFilter1D constructor: "
            "invalid length of the weight array (must be odd integer)");
        const unsigned n_half = n_kernel/2;

        // Copy the taper coefficients
        taper_ = new double[maxDegree + 1];
        if (itaper)
            copyBuffer(taper_, itaper, maxDegree+1);
        else
            for (unsigned i=0; i<=maxDegree; ++i)
                taper_[i] = 1.0;

        // Go over all bins and figure out the poly set
        PolyFilter1D *central = 0, *fitter = 0;
        bins_ = new PolyFilter1D*[dataLen];
        for (unsigned i=0; i<dataLen; ++i)
        {
            if (i < n_half || i + n_half >= dataLen)
            {
                // Create boundary weight and polynomials
                fitter = filterBuilder.makeFilter(taper_, maxDegree_,
                                                  i, dataLen);
                unique_.push_back(fitter);
            }
            else
            {
                // (Re)use the symmetric weight
                if (central == 0)
                {
                    // We should never get here if dataLen < n_kernel
                    assert(dataLen >= n_kernel);
                    central = filterBuilder.makeFilter(taper_, maxDegree_,
                                                       i, dataLen);
                    unique_.push_back(central);
                }
                fitter = central;
            }
            bins_[i] = fitter;
        }
    }

    LocalPolyFilter1D::~LocalPolyFilter1D()
    {
        for (unsigned i=unique_.size(); i>0; --i)
            delete unique_[i-1];
        delete [] bins_;
        delete [] taper_;
    }

    double LocalPolyFilter1D::taper(const unsigned degree) const
    {
        if (degree <= maxDegree_)
            return taper_[degree];
        else
            return 0.0;
    }

    const PolyFilter1D& LocalPolyFilter1D::getFilter(
        const unsigned binNumber)  const
    {
        if (binNumber >= nbins_) throw std::out_of_range(
            "In npstat::LocalPolyFilter1D::getFilter: bin number out of range");
        return *bins_[binNumber];
    }

    double LocalPolyFilter1D::selfContribution(const unsigned binNumber) const
    {
        if (binNumber >= nbins_) throw std::out_of_range(
            "In npstat::LocalPolyFilter1D::selfContribution: "
            "bin number out of range");
        const unsigned filterPeak = bins_[binNumber]->peakPosition();
        return bins_[binNumber]->at(filterPeak);
    }

    void LocalPolyFilter1D::clearSumBuffer() const
    {
        if (sumBuffer_.size() != nbins_)
            sumBuffer_.resize(nbins_);
        long double* sum = &sumBuffer_[0];
        for (unsigned i=0; i<nbins_; ++i)
            sum[i] = 0.0L;
    }

    void LocalPolyFilter1D::addWeightedFilter(const long double w,
                                              const unsigned binNum) const
    {
        const PolyFilter1D* filter(bins_[binNum]);
        const unsigned filterLen = filter->size();
        assert(filterLen <= nbins_);

        long double* sum = &sumBuffer_[0];
        const PolyFilter1D::value_type *f = &(*filter)[0];

        if (filterLen != nbins_)
        {
            if (filterLen % 2)
            {
                // Odd filter size. Are we in the center?
                const unsigned peak = filter->peakPosition();
                const unsigned center = filterLen/2;
                if (peak == center)
                {
                    // We are in the center
                    assert(binNum >= center);
                    sum += (binNum - center);
                }
                else if (peak > center)
                    // We are on the right side
                    sum += (nbins_ - filterLen);
            }
            else
            {
                // Even size filter. This can't be in the center.
                if (!(filter->peakPosition() < filterLen/2))
                    // We are on the right side
                    sum += (nbins_ - filterLen);
            }
        }

        for (unsigned i=0; i<filterLen; ++i)
            sum[i] += w*f[i];
    }

    CPP11_auto_ptr<LocalPolyFilter1D> symbetaLOrPEFilter1D(
        const int power, const double bandwidth, const unsigned degree,
        const unsigned grid_points, double interval_min,
        double interval_max, const bool stretchFilterAtTheBoundary)
    {
        if (grid_points == 0) throw std::invalid_argument(
            "In npstat::symbetaLOrPEFilter: number of grid points "
            "must be positive");

        if (bandwidth <= 0.0) throw std::invalid_argument(
            "In npstat::symbetaLOrPEFilter: bandwidth must be positive");

        CPP11_auto_ptr<AbsDistribution1D> kernel;
        if (power < 0)
            kernel = CPP11_auto_ptr<AbsDistribution1D>(
                new TruncatedGauss1D(0.0, bandwidth, 12.0));
        else
            kernel = CPP11_auto_ptr<AbsDistribution1D>(
                new SymmetricBeta1D(0.0, bandwidth, power));
        if (interval_min > interval_max)
            std::swap(interval_min, interval_max);
        const double grid_step = (interval_max - interval_min)/grid_points;

        CPP11_auto_ptr<LocalPolyFilter1D> result;
        if (stretchFilterAtTheBoundary)
        {
            StretchingFilter1DBuilder filterBuilder(kernel.get(), grid_step);
            result = CPP11_auto_ptr<LocalPolyFilter1D>(
                new LocalPolyFilter1D(0, degree, filterBuilder, grid_points));
        }
        else
        {
            std::vector<double> kernelScan;
            for (unsigned i=0; i<grid_points; ++i)
            {
                const double kernelValue = kernel->density(i*grid_step);
                if (kernelValue > 0.0)
                    kernelScan.push_back(kernelValue);
                else
                    break;
            }
            // Should add at least the point for i=0 to the scan
            assert(!kernelScan.empty());
            WeightTableFilter1DBuilder filterBuilder(
                &kernelScan[0], kernelScan.size());
            result = CPP11_auto_ptr<LocalPolyFilter1D>(
                new LocalPolyFilter1D(0, degree, filterBuilder, grid_points));
        }
        return result;
    }
}
