#ifndef NPSTAT_LOCALPOLYFILTER1D_HH_
#define NPSTAT_LOCALPOLYFILTER1D_HH_

/*!
// \file LocalPolyFilter1D.hh
//
// \brief Local polynomial filtering (regression) on 1-d equidistant grids
//
// This technique generalizes the Savitzky-Golay smoothing filter to
// arbitrary weight functions (not just a flat weight which leads to
// Legendre polynomials).
//
// Author: I. Volobouev
//
// November 2009
*/

#include "geners/CPP11_auto_ptr.hh"

#include "npstat/stat/AbsPolyFilter1D.hh"
#include "npstat/stat/AbsFilter1DBuilder.hh"

namespace npstat {
    /** This class performs local polynomial filtering in one dimension */
    class LocalPolyFilter1D : public AbsPolyFilter1D
    {
    public:
        /**
        // Main constructor. The arguments are as follows:
        //
        //  taper     -- Damping factors for each polynomial degree
        //               (starting with the 0th order term). This can be
        //               NULL in which case it is assumed that all
        //               factors are 1.
        //
        //  maxDegree -- Maximum degree of the polynomials. The length
        //               of the "taper" array (if not NULL) must be equal
        //               to maxDegree + 1. Note that, far away from the
        //               boundaries (where the situation is symmetric)
        //               the same filter will be produced using the same
        //               taper with an even degree N and with an odd degree
        //               N+1. Near the boundaries the filter coefficients
        //               will, of course, differ in these two cases.
        //
        //  filterBuilder -- An instance of a class which actually builds
        //                   the filters when requested. This builder
        //                   is used only inside the constructor.
        //
        //  dataLen   -- The length of the data arrays which will be
        //               used with this filter (this info is needed in
        //               order to take into account the boundary effects).
        */
        LocalPolyFilter1D(const double* taper, unsigned maxDegree,
                          const AbsFilter1DBuilder& filterBuilder,
                          unsigned dataLen);
        virtual ~LocalPolyFilter1D();

        //@{
        /** Inspect object properties */
        double taper(unsigned degree) const;
        inline unsigned maxDegree() const {return maxDegree_;}
        inline unsigned dataLen() const {return nbins_;}
        //@}

        /** Self contribution needed for cross-validation */
        double selfContribution(unsigned binNumber) const;

        /** Get the filter coefficients for the given bin */
        const PolyFilter1D& getFilter(unsigned binNumber) const;

        /**
        // This method performs the filtering. "dataLen",
        // which is the length of both "in" and "out" arrays,
        // must be the same as the one in the constructor.
        */
        template <typename Tin, typename Tout>
        void filter(const Tin* in, unsigned dataLen, Tout* out) const;

        /**
        // A diffent filtering method in which the shapes of the
        // kernels are determined by the positions of the "sources"
        // (i.e., sample points) instead of the positions at which
        // the density (or response) is estimated.
        */
        template <typename Tin, typename Tout>
        void convolve(const Tin* in, unsigned dataLen, Tout* out) const;

    private:
        LocalPolyFilter1D();
        LocalPolyFilter1D(const LocalPolyFilter1D&);
        LocalPolyFilter1D& operator=(const LocalPolyFilter1D&);

        mutable std::vector<long double> sumBuffer_;
        std::vector<PolyFilter1D*> unique_;
        double* taper_;
        PolyFilter1D** bins_;
        unsigned nbins_;
        unsigned maxDegree_;

        template <typename T>
        double convolute(const T* data, const PolyFilter1D* filter) const;

        void addWeightedFilter(long double w, unsigned binNum) const;

        void clearSumBuffer() const;
    };

    /**
    // The utility function that generates the most common filters using
    // kernels from the symmetric beta family. The arguments are as follows:
    //
    //  m          -- Choose the kernel from the symmetric beta family
    //                proportional to (1 - x^2)^m. If m is negative,
    //                Gaussian kernel will be used, truncated at +- 12 sigma.
    //
    //  bandwidth  -- Kernel bandwidth.
    //
    //  maxDegree  -- Degree of the LOrPE polynomial (as in the
    //                LocalPolyFilter1D constructor).
    //
    //  numberOfGridPoints -- Length of the data array to be used with this
    //                filter (typically, number of histogram bins and such)
    //
    //  xmin, xmax -- Data grid limits (as in a histogram).
    //
    //  stretchFilterAtTheBoundary -- If "true", try to stretch the filter
    //                bandwidth at the boundary. This will increase bias and
    //                decrease variance near the density support boundaries.
    */
    CPP11_auto_ptr<LocalPolyFilter1D> symbetaLOrPEFilter1D(
        int m, double bandwidth, unsigned maxDegree,
        unsigned numberOfGridPoints, double xmin, double xmax,
        bool stretchFilterAtTheBoundary=false);

#ifdef SWIG
    LocalPolyFilter1D* symbetaLOrPEFilter1D_2(
        int m, double bandwidth, unsigned maxDegree,
        unsigned numberOfGridPoints, double xmin, double xmax,
        bool stretchFilterAtTheBoundary=false);
#endif
}

#include "npstat/stat/LocalPolyFilter1D.icc"

#endif // NPSTAT_LOCALPOLYFILTER1D_HH_
