#ifndef NPSTAT_NONPARAMETRICCOMPOSITEBUILDER_HH_
#define NPSTAT_NONPARAMETRICCOMPOSITEBUILDER_HH_

/*!
// \file NonparametricCompositeBuilder.hh
//
// \brief Density estimation decomposed: do it separately for the marginals
//        and the copula
//
// Author: I. Volobouev
//
// September 2010
*/

#include "npstat/stat/AbsCompositeDistroBuilder.hh"
#include "npstat/stat/AbsMarginalSmoother.hh"
#include "npstat/stat/AbsCopulaSmoother.hh"

namespace npstat {
    /**
    // A class for building nonparametric estimates of multivariate
    // densities by performing separate density estimation for each
    // marginal and the copula. This allows us to introduce a larger
    // number of bandwidth parameters into the system which can
    // often lead to a better estimate.
    */
    template <class Point>
    class NonparametricCompositeBuilder : 
        public AbsCompositeDistroBuilder<Point>
    {
    public:
        /**
        // Constructor arguments are as follows:
        //
        // smoothND              -- Copula smoother pointer
        //
        // smooth1D              -- Array of pointers for margin smoothers
        //
        // takePointerOwnership  -- if true, delete smoothND and pointers
        //                          inside smooth1D in the destructor
        //
        // interpolationDegreeCopula    -- interpolation degree for the
        //                                 copula density array. Currently
        //                                 must be 0 or 1, limited by the
        //                                 abilities of BinnedDensityND
        //                                 class.
        //
        // interpolationDegreeMarginals -- interpolation degree for the
        //                                 marginal density array. Currently
        //                                 must be 0 or 1, limited by the
        //                                 abilities of BinnedDensity1D
        //                                 class.
        */
        NonparametricCompositeBuilder(
            AbsCopulaSmoother* smoothND,
            const std::vector<AbsMarginalSmoother*>& smooth1D,
            bool takePointerOwnership,
            unsigned interpolationDegreeCopula,
            unsigned interpolationDegreeMarginals);

        virtual ~NonparametricCompositeBuilder();

        /** Bandwidth factor used in the most recent copula smoothing */
        inline double lastCopulaBandwidth() const
            {return copulaBandwidth_;}

        /** Bandwidth used in the most recent smoothing of the marginals */
        inline double lastMarginBandwidth(const unsigned i) const
            {return marginBandwidth_.at(i);}

        /** Dimensionality of the reconstructed density */
        inline unsigned dim() const {return smooth1D_.size();}

    private:
        NonparametricCompositeBuilder();
        NonparametricCompositeBuilder(const NonparametricCompositeBuilder&);
        NonparametricCompositeBuilder& operator=(
            const NonparametricCompositeBuilder&);

        virtual AbsDistribution1D* buildMarginal(
            unsigned dimNumber,
            const std::vector<typename Point::value_type>&) const;

        virtual AbsDistributionND* buildCopula(
            std::vector<OrderedPointND<Point> >&) const;

        AbsCopulaSmoother* smoothND_;
        std::vector<AbsMarginalSmoother*> smooth1D_;
        mutable std::vector<double> marginBandwidth_;
        mutable double copulaBandwidth_;
        unsigned iDegCopula_;
        unsigned iDegMarginals_;
        bool takePointerOwnership_;
    };
}

#include "npstat/stat/NonparametricCompositeBuilder.icc"

#endif // NPSTAT_NONPARAMETRICCOMPOSITEBUILDER_HH_
