#include <utility>

#include "npstat/stat/PolyFilterCollection1D.hh"
#include "npstat/stat/Filter1DBuilders.hh"

namespace npstat {
    PolyFilterCollection1D::PolyFilterCollection1D(
        const AbsScalableDistribution1D& kernel,
        const double* taper, const unsigned maxDegree,
        const double binwidth, const unsigned dataLen,
        const bool stretchKernelAtBoundaries)
        : scanBuf_(dataLen),
          kernel_(kernel.clone()),
          binwidth_(binwidth),
          taper_(0),
          maxDegree_(maxDegree),
          dataLen_(dataLen),
          stretchKernel_(stretchKernelAtBoundaries)
    {
        if (binwidth <= 0.0) throw std::invalid_argument(
            "In npstat::PolyFilterCollection1D constructor: "
            "bin width must be positive");
        if (!dataLen) throw std::invalid_argument(
            "In npstat::PolyFilterCollection1D constructor: "
            "expected data length must be positive");
        if (taper)
        {
            taper_ = new double[maxDegree + 1];
            for (unsigned i=0; i<=maxDegree; ++i)
                taper_[i] = taper[i];
        }
        kernel_->setLocation(0.0);
    }

    PolyFilterCollection1D::~PolyFilterCollection1D()
    {
        delete kernel_;
        delete [] taper_;
        for (std::map<double,const LocalPolyFilter1D*>::iterator it =
                 filterMap_.begin(); it != filterMap_.end(); ++it)
            delete it->second;
    }

    void PolyFilterCollection1D::processBandwidth(const double bandwidth)
    {
        kernel_->setScale(bandwidth);

        // Construct filter builder
        AbsFilter1DBuilder* builder = 0;
        if (stretchKernel_)
            builder = new StretchingFilter1DBuilder(kernel_, binwidth_);
        else
        {
            // We need to scan the kernel by hand. Assume that the weight
            // is symmetric.
            double* scan = &scanBuf_[0];
            unsigned imax = 0;
            for (; imax<dataLen_; ++imax)
            {
                scan[imax] = kernel_->density(imax*binwidth_);
                if (scan[imax] <= 0.0)
                    break;
            }
            builder = new WeightTableFilter1DBuilder(scan, imax);
        }

        // Construct the filter
        LocalPolyFilter1D* filter = new LocalPolyFilter1D(
            taper_, maxDegree_, *builder, dataLen_);
        filterMap_.insert(std::make_pair(bandwidth, filter));

        delete builder;
    }

    double PolyFilterCollection1D::taper(const unsigned i) const
    {
        if (i > maxDegree_)
            return 0.0;
        if (taper_)
            return taper_[i];
        else
            return 1.0;
    }
}
