#ifndef NPSTAT_AMISEOPTIMALBANDWIDTH_HH_
#define NPSTAT_AMISEOPTIMALBANDWIDTH_HH_

/*!
// \file amiseOptimalBandwidth.hh
//
// \brief Optimal AMISE bandwidth for KDE with high order kernels
//
// AMISE stands for Asymptotic Mean Integrated Squared Error.
// The formulae used come from the paper "Bandwidth Selection
// in Kernel Density Estimation: A Review" by B.A. Turlach.
//
// Author: I. Volobouev
//
// July 2010
*/

namespace npstat {
    /**
    // Estimate optimal bandwidth for filters generated by the Gaussian
    // distribution used as the weight function. The arguments are as follows:
    //
    // filterDegree --  Degree of the filter (in the LOrPE sense). This
    //                  degree is less by 2 then the order of the kernel
    //                  in the Turlach's paper. For example, filter of
    //                  degree 0 corresponds to kernel of order 2 (original
    //                  Gaussian), filter of degree 0 corresponds to kernel
    //                  of order 4, etc. Here, the degree must be even.
    //                  Maximum possible filter degree is 42.
    //
    // npoints      --  Number of points in the data sample.
    //
    // fvalues      --  Array of scanned values of the reference density.
    //                  Note that this table is NOT preserved.
    //
    // arrLen       --  Number of elements in the array "fvalues".
    //
    // h            --  Step with which the scan of the reference density
    //                  was performed.
    //
    // expectedAmise -- If this argument is provided, it will be filled
    //                  with the expected AMISE value.
    //
    // The formulae used involve numerical evaluations of the derivatives
    // of the scanned reference density. For a given filterDegree, the
    // code needs to calculate the derivative of order filterDegree + 2.
    // Make sure the scan step is chosen appropriately for this kind of
    // calculation (see, for example, the "Numerical Recipes" book in order
    // to understand the issues involved).
    */
    template<typename Real>
    double amiseOptimalBwGauss(unsigned filterDegree, double npoints,
                               Real* fvalues, unsigned long arrLen, Real h,
                               double* expectedAmise = 0);

    /**
    // Estimate optimal bandwidth for filters generated by the symmetric
    // beta distribution used as the weight function. The "power" argument
    // is the power of the symmetric beta, limited here to unsigned integers.
    // Maximum possible power value is 10. The meaning of all other arguments
    // is the same as in the function npstat::amiseOptimalBwGauss
    */
    template<typename Real>
    double amiseOptimalBwSymbeta(unsigned power, unsigned filterDegree,
                                 double npoints,
                                 Real* fvalues, unsigned long arrLen, Real h,
                                 double* expectedAmise = 0);

    /**
    // Plug-in bandwidth for filters generated by the Gaussian distribution
    // used as the weight function
    */
    double amisePluginBwGauss(unsigned filterDegree, double npoints,
                              double sampleSigma, double* expectedAmise = 0);

    /**
    // Plug-in bandwidth for filters generated by the symmetric beta
    // distribution used as the weight function. The "power" argument
    // is the power of the symmetric beta, limited here to unsigned integers.
    // Maximum possible power value is 10.
    */
    double amisePluginBwSymbeta(unsigned power, unsigned filterDegree,
                                double npoints, double sampleSigma,
                                double* expectedAmise = 0);
}

#include "npstat/stat/amiseOptimalBandwidth.icc"

#endif // NPSTAT_AMISEOPTIMALBANDWIDTH_HH_
